(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: identity_model.ml,v 1.2 2014-03-11 09:53:15 fclement Exp $ *)

module Nat = Unatural;;
module Cmd_opt = Ucommand_option;;

open Worker;;

(* Abtract references. *)

let init_dim = Nat._0_;;

let { Uaref.get = get_dim; set = set_dim; } =
  Uaref.make_aref_unique "dimension" init_dim
;;

let init_data_image = Umatrix.of_array_array [||];;

let { Uaref.get = get_data_image; set = set_data_image; } =
  Uaref.make_aref_unique "data image" init_data_image
;;

let init_result_image_file_name = "";;

let { Uaref.get = get_result_image_file_name;
      set = set_result_image_file_name; } =
  Uaref.make_aref_unique "result image file name"
    init_result_image_file_name
;;

(* Image stuff. *)

type image = Colors.t Umatrix.t;;

let sprint_color_matrix = sprint_matrix Colors_print.print_t;;

let load_data_image fname =
  let dim =
    match Ppm.ppm_magic_number fname with
    | Ppm.P1 | Ppm.P2 | Ppm.P4 | Ppm.P5 -> Nat._1_
    | Ppm.P3 | Ppm.P6 -> Nat._3_ in
  set_dim dim;
  let pixmap = Ppm.load_pixmap fname in
  let idata = Umatrix.of_array_array pixmap in
  let (h, w) = Umatrix.size idata in
  if h = Nat._0_ || w = Nat._0_ then
    um_invalid_arg "load_data_image" "empty image" else
  set_data_image idata;
;;

let float_triple_of_color c =
  let (r, g, b) = (Colors.r_of c, Colors.g_of c, Colors.b_of c) in
  (float_of_int r, float_of_int g, float_of_int b)
;;

let of_color dim c =
  let (r, g, b) = float_triple_of_color c in
  if dim = Nat._1_ then
    begin
      if r = g && g = b then [| r; |] else
      um_failwith "of_color" "color is not grayscale"
    end
  else if dim = Nat._3_ then [| r; g; b; |]
  else assert false
;;

let data_of_image dim img =
  let im = Umatrix.map (of_color dim) img in
  Umatrix.transpose (Umatrix.of_array_array (Umatrix.to_array im))
;;

let float_triple_to_color (r, g, b) =
  Colors.color_of_rgb (int_of_float r) (int_of_float g) (int_of_float b)
;;

let to_color dim a =
  if Uarray.length a <> dim then
    um_invalid_arg "to_color" "wrong dimension";
  let (r, g, b) =
    if dim = Nat._1_ then (a.(0), a.(0), a.(0)) else
    if dim = Nat._3_ then (a.(0), a.(1), a.(2)) else
    assert false in
  float_triple_to_color (r, g, b)
;;

let image_of_fine_parameter dim (ni, nj) pm =
  let im =
    Uarray_ng.matrix_of_array ni nj
      (Umatrix.to_array_array (Umatrix.transpose pm)) in
  Umatrix.map (to_color dim) im
;;

(* Command-line options. *)

let option_result_image_file_name =
  {
   Cmd_opt.key = "-result-parameter-file-name";
   spec = Arg.String set_result_image_file_name;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet result parameter file name to <fname>.";
  }
;;

let regular_opt rco =
  Cmd_opt.with_auto_short_key
    (Cmd_opt.with_auto_gnu_key
       (Cmd_opt.of_regular rco))
;;

let user_add x = Cmd_opt.user_add (regular_opt x);;

let add_options () = user_add option_result_image_file_name;;

add_options ();;

Cmd_opt.set_usage_msg
  (Format.sprintf
     "@.%s -- Identity worker@.@.\
      Usage: %s [<option>...] <input_image_file_name>@.@.\
      Options are:@."
     Sys.argv.(0) Sys.argv.(0))
;;

Cmd_opt.set_anon_fun load_data_image;;

(* Computational server. *)

let init_data = Umatrix.make Nat._1_ Nat._1_ 0.
and init_data_width = Nat._1_
and init_data_height = Nat._1_
;;
let { Uaref.get = get_data; set = set_data; } =
  Uaref.make_aref_unique "data" init_data
and { Uaref.get = get_data_width; set = set_data_width; } =
  Uaref.make_aref_unique "data width" init_data_width
and { Uaref.get = get_data_height; set = set_data_height; } =
  Uaref.make_aref_unique "data height" init_data_height
;;

let set_data_and_size dm (ni, nj) =
  set_data_height ni;
  set_data_width nj;
  set_data dm;
;;

let string_starts_with sub s = Ustring.sub s 0 (Ustring.length sub) = sub;;

let check_initialized tname f x =
  let y =
    try f x with
    | Failure s when string_starts_with "Fatal error in Uaref" s ->
        raise (Pio.Uninitialized_worker tname) in
  y
;;

(* Optim. *)

let add_error2 s x y = s +. (x -. y)**2.;;

let cost dm ppm maa =
  let pm = coarse_parameter_to_fine_parameter ppm maa in
  0.5 *. (Umatrix.fold_left2 add_error2 0. dm pm)
;;

let optim_engine _spf ppm maa_init =
  let maa_opt = Uarray_ng.array_array_map (fun _ -> 0.) maa_init
  and count = Uarray_ng.array_array_map (fun _ -> ref 0) maa_init in
  let dm = check_initialized "optim" get_data () in
  if Umatrix.size dm <> Umatrix.size ppm then assert false else
  let add i j zn =
    let zn = Unatural.to_int zn in
    maa_opt.(i).(zn) <- maa_opt.(i).(zn) +. Umatrix.get dm i j;
    incr count.(i).(zn); in
  Umatrix.iter_ij add ppm;
  let maa_opt =
    Uarray_ng.array_array_map2
      (fun mi ci -> mi /. float_of_int !ci)
      maa_opt count in
  let j_opt = cost dm ppm maa_opt in
  (maa_opt, j_opt)
;;

let optim = make_optim optim_engine;;

(* Init. *)

let init_engine { Udebug.spf_debug = debug; _; } dim =
  if dim <> Nat._1_ && dim <> Nat._3_ then
    um_invalid_arg "init" "dimension must be 1 or 3";
  let dim' = get_dim () in
  check_valid_dimension "init" dim dim';
  let img = get_data_image () in
  debug (sprint_color_matrix "img" img);
  let sz = Umatrix.size img in
  let dm = data_of_image dim img in
  let nbpt = Umatrix.dim_j dm in
  debug (sprint_float_matrix "dm" dm);
  set_data_and_size dm sz;
  let ppm1 = Umatrix.make dim nbpt Unatural._0_ in
  let maa1_init = Uarray.init dim (fun _i -> Uarray.make Unatural._1_ 0.) in
  (sz, ppm1, maa1_init)
;;

let init = make_init init_engine optim;;

(* Grad. *)

let grad_engine { Udebug.spf_debug = debug; _; } pm =
  let dm = check_initialized "grad" get_data () in
  debug (sprint_float_matrix "dm" dm);
  Umatrix.map2 ( -. ) pm dm
;;

let grad = make_grad grad_engine;;

(* Final. *)

let final_engine _spf pm =
  let ni = check_initialized "final" get_data_height ()
  and nj = check_initialized "final" get_data_width () in
  let img = image_of_fine_parameter (get_dim ()) (ni, nj) pm in
  let fname = get_result_image_file_name () in
  Ppm.save_pixmap fname (Umatrix.to_array_array img);
;;

let final = make_final final_engine ;;
